/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * api.c:  The Solver's API wrappings for various optimization algorithms.
 *
 * Author:
 *   Jukka-Pekka Iivonen <jiivonen@hutcs.cs.hut.fi>
 *
 * (C) Copyright 2002 by Jukka-Pekka Iivonen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <gnumeric-config.h>
#include <glib/gi18n-lib.h>
#include "gnumeric.h"
#include "numbers.h"

#include "solver.h"
#include "func.h"
#include "cell.h"
#include "sheet.h"
#include "sheet-style.h"
#include "dependent.h"
#include "dialogs.h"
#include "mstyle.h"
#include "value.h"
#include "mathfunc.h"
#include "analysis-tools.h"

#include <math.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "lp_solve/lp_solve.h"
#include "api.h"
#include "glpk.h"

#include "coplqp/QPdef.h"
#include "ooqp/ooqp.h"
#include "qld/qld.h"


/* ------------------------------------------------------------------------- */

/*
 * Solver's API wrappings for the LP Solve 5.5.
 *
 * Package:    LP Solve
 * Version:    5.5
 * License:    LGPL
 * Homepage:   
 */

typedef struct {
        lprec    *p;
        gboolean assume_non_negative;
} lp_solve_t;

static void
w_lp_set_obj_mat_fn_dummy (SolverProgram program, int col, int row,
			   gnm_float value)
{
	printf(" w_lp_set_obj_mat_fn_dummy \n");
}

static SolverProgram
w_lp_solve_init (const SolverParameters *param)
{
	lp_solve_t *lp;

	lp                      = g_new (lp_solve_t, 1);
	lp->assume_non_negative = param->options.assume_non_negative;
	if (lp->assume_non_negative)
	        lp->p = lp_solve_make_lp (param->n_constraints,
					  param->n_variables);
	else
	        lp->p = lp_solve_make_lp (param->n_constraints,
					  2 * param->n_variables);

	return lp;
}

static void
w_lp_solve_delete_lp (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;

        lp_solve_delete_lp (lp->p);
	g_free (lp);
}

static void
w_lp_solve_set_maxim (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;

        lp_solve_set_maxim (lp->p);
}

static void
w_lp_solve_set_minim (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;

        lp_solve_set_minim (lp->p);
}

static void
w_lp_solve_set_obj_fn (SolverProgram program, int col, gnm_float value)
{
	lp_solve_t *lp = (lp_solve_t *) program;

	if (lp->assume_non_negative)
	        lp_solve_set_mat (lp->p, 0, col + 1, value);
	else {
	        lp_solve_set_mat (lp->p, 0, 2 * col + 1, value);
	        lp_solve_set_mat (lp->p, 0, 2 * col + 2, -value);
	}
}

static void
w_lp_solve_set_constr_mat (SolverProgram program, int col, int row,
			   gnm_float value)
{
	lp_solve_t *lp = (lp_solve_t *) program;

	if (lp->assume_non_negative)
	        lp_solve_set_mat (lp->p, row + 1, col + 1, value);
	else {
	        lp_solve_set_mat (lp->p, row + 1, 2 * col + 1, value);
	        lp_solve_set_mat (lp->p, row + 1, 2 * col + 2, -value);
	}
}

static void
w_lp_solve_set_constr (SolverProgram program, int row,
		       SolverConstraintType type, gnm_float rhs)
{
	lp_solve_t *lp = (lp_solve_t *) program;
	int lp_constraint_type;
	switch (type) {
	case SolverLE:	lp_constraint_type = ROWTYPE_LE; break;
	case SolverGE:	lp_constraint_type = ROWTYPE_GE; break;
	case SolverEQ:	lp_constraint_type = ROWTYPE_EQ; break;
	default:
		g_warning ("unexpected constraint type %d", type);
		lp_constraint_type = 0; /* silence the compiler */
	}
        lp_solve_set_constr_type (lp->p, row + 1, lp_constraint_type);
        lp_solve_set_rh (lp->p, row + 1, rhs);
}

static void
w_lp_solve_set_int (SolverProgram program, int col)
{
	lp_solve_t *lp = (lp_solve_t *) program;

	if (lp->assume_non_negative) 
	        lp_solve_set_int (lp->p, col + 1, TRUE);
	else {
	        lp_solve_set_int (lp->p, 2 * col + 1, TRUE);
	        lp_solve_set_int (lp->p, 2 * col + 2, TRUE);
	}
}

static void
w_lp_solve_set_bool (SolverProgram program, int col)
{
	lp_solve_t *lp = (lp_solve_t *) program;

	if (lp->assume_non_negative) {
	        lp_solve_set_int   (lp->p, col + 1, TRUE);
		lp_solve_set_upbo  (lp->p, col + 1, 1);
		lp_solve_set_lowbo (lp->p, col + 1, 0);
	} else {
	        lp_solve_set_int   (lp->p, 2 * col + 1, TRUE);
		lp_solve_set_upbo  (lp->p, 2 * col + 1, 1);
		lp_solve_set_lowbo (lp->p, 2 * col + 1, 0);

	        lp_solve_set_int   (lp->p, 2 * col + 2, TRUE);
		lp_solve_set_upbo  (lp->p, 2 * col + 2, 0);
		lp_solve_set_lowbo (lp->p, 2 * col + 2, 0);
	}
}

static void
w_lp_solve_print_lp (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;

	lp_solve_print_lp (lp->p);
}

static SolverStatus
w_lp_solve_solve (SolverProgram program)
{
	int res;
	lp_solve_t *lp = (lp_solve_t *) program;

#if SOLVER_DEBUG
	w_lp_solve_print_lp (program);
#endif
        switch ((res = lp_solve_solve (lp->p))) {
	default:
		g_warning ("unknown result from lp_solve_solve '%d'" , res);

	case UNKNOWNERROR:
	case DATAIGNORED:
	case NOBFP:
	case NOMEMORY:
	case NOTRUN:
	case DEGENERATE:
	case NUMFAILURE:
	case USERABORT:
	case TIMEOUT:
		return SolverFailure;

	case OPTIMAL:	return SolverOptimal;
	case SUBOPTIMAL: return SolverMaxIterExc; /* or SolverMaxTimeExc */
	case INFEASIBLE:return SolverInfeasible;
	case UNBOUNDED:	return SolverUnbounded;
	case RUNNING:	return SolverRunning;
	}
}

static gnm_float
w_lp_solve_get_solution (SolverProgram program, int column)
{
	lp_solve_t *lp = (lp_solve_t *) program;
	int nrows = lp_solve_get_nrows (lp->p);

	if (lp->assume_non_negative)
		return lp_solve_get_primal (lp->p, nrows + column + 1);
	else {
		int i = nrows + 2 * column + 1;
	        gnm_float x = lp_solve_get_primal (lp->p, i);
	        gnm_float neg_x = lp_solve_get_primal (lp->p, i + 1);

		if (x > neg_x)
		        return x;
		else
		        return -neg_x;
	}
}

static gnm_float
w_lp_solve_get_value_of_obj_fn (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;
	g_return_val_if_fail (lp != NULL, 0.);
	g_return_val_if_fail (lp->p != NULL, 0.);

	return lp_solve_get_primal (lp->p, 0);
}

static gnm_float
w_lp_solve_get_dual (SolverProgram program, int row)
{
	lp_solve_t *lp = (lp_solve_t *) program;
	g_return_val_if_fail (lp != NULL, 0.);
	g_return_val_if_fail (lp->p != NULL, 0.);

	return lp_solve_get_dual (lp->p, row + 1);
}

static int
w_lp_solve_get_iterations (SolverProgram program)
{
	lp_solve_t *lp = (lp_solve_t *) program;
	g_return_val_if_fail (lp != NULL, 0);
	g_return_val_if_fail (lp->p != NULL, 0);

	/* FIXME: gint64 */
        return lp_solve_get_total_iter (lp->p);
}

static gboolean
w_lp_solve_set_option (SolverProgram program, SolverOptionType option,
		       const gboolean *b_value,
		       const gnm_float *f_value, const int *i_value)
{
	lp_solve_t *lp = (lp_solve_t *) program;

        switch (option) {
	case SolverOptAutomaticScaling:
#if 0
	No longer needed with 5.x
#endif
	        return FALSE;
	case SolverOptMaxIter:
	        lp_solve_set_scalelimit (lp->p, *i_value);
	        return FALSE;
	case SolverOptMaxTimeSec:
	        lp_solve_set_timeout (lp->p, *i_value);
	        return FALSE;
	default:
	        return TRUE;
	}
}


/* ------------------------------------------------------------------------- */

/*
 * Solver's API wrappings for the GLPK 3.2.
 *
 * Package:    GLPK
 * Version:    3.2 (Jul 15, 2002)
 * License:    GPL
 * Homepage:   http://www.gnu.org/software/glpk/glpk.html
 * Algorithm:  revised simplex method
 *
 */


typedef struct {
        LPX         *p;
	int         *rn;
	int         *cn;
	gnm_float  *a;
	int         n;
        gboolean    assume_non_negative;
        gboolean    scaling;
} glpk_simplex_t;

static SolverProgram
w_glpk_init (const SolverParameters *param)
{
        glpk_simplex_t *lp;
	int             i, cols;

	lp                      = g_new (glpk_simplex_t, 1);
	lp->p                   = lpx_create_prob ();
	lp->assume_non_negative = param->options.assume_non_negative;
	lp->scaling             = param->options.automatic_scaling;

	cols = param->n_variables;

	lpx_add_cols (lp->p, cols);
	lpx_add_rows (lp->p, param->n_constraints);
	lp->a  = g_new (gnm_float, cols * param->n_constraints + 1); 
	lp->cn = g_new (int, cols * param->n_constraints + 1); 
	lp->rn = g_new (int, cols * param->n_constraints + 1); 
	lp->n  = 1;

	if (lp->assume_non_negative)
		for (i = 0; i < cols; i++)
			lpx_set_col_bnds (lp->p, i + 1, LPX_LO, 0, 0);
	else
		for (i = 0; i < cols; i++)
			lpx_set_col_bnds (lp->p, i + 1, LPX_FR, 0, 0);

	return lp;
}

static void
w_glpk_delete_lp (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

        lpx_delete_prob (lp->p);
	g_free (lp->cn);
	g_free (lp->rn);
	g_free (lp->a);
	g_free (lp);
}

static void
w_glpk_set_maxim (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

        lpx_set_obj_dir (lp->p, LPX_MAX);
}

static void
w_glpk_set_minim (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

        lpx_set_obj_dir (lp->p, LPX_MIN);
}

static void
w_glpk_set_obj_fn (SolverProgram program, int col, gnm_float value)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

	lpx_set_obj_coef (lp->p, col + 1, value);
}

static void
w_glpk_set_constr_mat (SolverProgram program, int col, int row,
		       gnm_float value)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

	lp->cn [lp->n] = col + 1;
	lp->rn [lp->n] = row + 1;
	lp->a  [lp->n] = value;
	(lp->n)++;
}

static void
w_glpk_set_constr (SolverProgram program, int row, SolverConstraintType type,
		   gnm_float value)
{
        int            typemap [] = { LPX_UP, LPX_LO, LPX_FX, -1, -1, -1 };
        glpk_simplex_t *lp        = (glpk_simplex_t *) program;

	if (typemap [type] != -1)
	        lpx_set_row_bnds (lp->p, row + 1, typemap [type], value,
				  value);
	else
	        printf ("Error\n");
}

static void
w_glpk_set_int (SolverProgram program, int col)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

	lpx_set_class (lp->p, LPX_MIP);
	lpx_set_col_kind (lp->p, col + 1, LPX_IV);
}

static void
w_glpk_set_bool (SolverProgram program, int col)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

	lpx_set_class (lp->p, LPX_MIP);
	lpx_set_col_kind (lp->p, col + 1, LPX_IV);
	lpx_set_col_bnds (lp->p, col + 1, LPX_DB, 0, 1);
}

static void
w_glpk_print_lp (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;
        int            i, n, cols, rows;
	int            typex;
	gnm_float     lb, ub;

	cols = lpx_get_num_cols (lp->p);
	rows = lpx_get_num_rows (lp->p);

	printf ("\t\t");
	for (i = 0; i < cols; i++)
	        printf ("Var[%3d] ", i + 1);
	printf ("\n");

	if (lpx_get_obj_dir (lp->p) == LPX_MAX)
	        printf ("Maximize\t");
	else
	        printf ("Minimize\t");
	for (i = 0; i < cols; i++)
	        printf ("%8g ", lpx_get_obj_coef (lp->p, i + 1));
	printf ("\n");

	for (i = 0; i < rows; i++) {
		gnm_float *a;
		int        *ndx, t;

	        printf ("Row[%3d]\t", i + 1);

		a   = g_new (gnm_float, cols + 1);
		ndx = g_new (int, cols + 1);
		lpx_get_mat_row (lp->p, i + 1, ndx, a);
		for (n = 0, t = 1; n < cols; n++) {
			if (ndx [t] == n + 1)
				printf ("%8g ", a[t++]);
			else
				printf ("%8g ", 0.0);
		}
		g_free (ndx);
		g_free (a);
		
		lpx_get_row_bnds (lp->p, i + 1, &typex, &lb, &ub);
		if (typex == LPX_LO)
		        printf (">= %8g\n", lb);
		else if (typex == LPX_UP)
		        printf ("<= %8g\n", ub);
		else
		        printf ("=  %8g\n", lb);
	}

	printf ("Type\t\t");
	for (i = 0; i < cols; i++)
		if (lpx_get_class (lp->p) == LPX_LP 
		    || lpx_get_col_kind (lp->p, i + 1) == LPX_CV)
			printf ("  Real\t");
		else
			printf ("  Int\t");

	printf ("\nupbo\t\t");
	for (i = 0; i < cols; i++) {
		lpx_get_col_bnds (lp->p, i + 1, &typex, &lb, &ub);
		if (typex == LPX_LO || typex == LPX_FR)
			printf ("Infinite  ");
		else
			printf ("%8g ", ub);
	}

	printf ("\nlowbo\t\t");
	for (i = 0; i < cols; i++) {
		lpx_get_col_bnds (lp->p, i + 1, &typex, &lb, &ub);
		if (typex == LPX_UP || typex == LPX_FR)
			printf ("-Infinite ");
		else
			printf ("%8g ", ub);
	}
	printf ("\n");
}

static SolverStatus
w_glpk_simplex_solve (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;
	
	if (lp->scaling)
	        lpx_scale_prob (lp->p);

	lpx_load_matrix (lp->p, lp->n - 1, lp->rn, lp->cn, lp->a);

#if SOLVER_DEBUG
	w_glpk_print_lp (program);
#endif

	lpx_simplex (lp->p);
	if (lpx_get_class (lp->p) == LPX_MIP) {
		switch (lpx_get_status (lp->p)) {
		case LPX_OPT:
			break;
		case LPX_INFEAS:
			return SolverInfeasible;
		case LPX_UNBND:
			return SolverUnbounded;
		default:
			printf ("Error: w_glpk_simplex_solve\n");
			return SolverInfeasible;
		}

		lpx_intopt (lp->p);
		switch (lpx_mip_status (lp->p)) {
		case LPX_I_OPT:
			return SolverOptimal;
		case LPX_I_NOFEAS:
			return SolverInfeasible;
		default:
			printf ("Error: w_glpk_simplex_solve\n");
			return SolverInfeasible;
		}
	} else {
		if (lp->scaling)
			lpx_unscale_prob (lp->p);

		switch (lpx_get_status (lp->p)) {
		case LPX_OPT:
			return SolverOptimal;
		case LPX_INFEAS:
			return SolverInfeasible;
		case LPX_UNBND:
			return SolverUnbounded;
		default:
			printf ("Error: w_glpk_simplex_solve\n");
			return SolverInfeasible;
		}
	}
}

static gnm_float
w_glpk_get_solution (SolverProgram program, int col)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;
        gnm_float      x;

        if (lpx_get_class (lp->p) == LPX_LP)
		lpx_get_col_info (lp->p, col + 1, NULL, &x, NULL);
	else
		x = lpx_mip_col_val (lp->p, col + 1);

	return x;
}

static gnm_float
w_glpk_get_value_of_obj_fn (SolverProgram program)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;

        if (lpx_get_class (lp->p) == LPX_LP)
		return lpx_get_obj_val (lp->p);
	else
		return lpx_mip_obj_val (lp->p);
}

static gnm_float
w_glpk_get_dual (SolverProgram program, int row)
{
        glpk_simplex_t *lp = (glpk_simplex_t *) program;
        gnm_float      x;

        lpx_get_row_info (lp->p, row + 1, NULL, NULL, &x);
	return x;
}

static int
w_glpk_get_iterations (SolverProgram program)
{
        return -1;
}

static gboolean
w_glpk_set_option (SolverProgram program, SolverOptionType option,
		   const gboolean *b_value,
		   const gnm_float *f_value, const int *i_value)
{
        switch (option) {
	case SolverOptAutomaticScaling:
	        return FALSE;
	case SolverOptMaxIter:
printf ("FIXME: Max iter=%d\n", *i_value);
	        return FALSE;
	case SolverOptMaxTimeSec:
printf ("FIXME: Max time (sec.)=%d\n", *i_value);
	        return FALSE;
	default:
	        return TRUE;
	}
}


/* ------------------------------------------------------------------------- */

/*
 * Solver's API wrappings for QP .
 *
 * Package:    QLD
 * Version:    0
 * License:    own
 * Homepage:   -
 * Algorithm:  -
 *
 */

static SolverProgram
w_qld_init (const SolverParameters *param)
{
	int bulharska_konstanta = 1;	//TODO - static
	qp_qld_program *p = g_new (qp_qld_program, 1);

        printf ("w_qld_init\n");
	assert(param->n_constraints == param->n_total_constraints);

	p->m = param->n_constraints;
	p->mmax = p->m + bulharska_konstanta;
	p->me = 0;
 	p->n = param->n_variables;
	p->nmax = p->n;
	p->mnn = p->m + p->n + p->n;

        p->lwar = 3 *p->n * p->n / 2    + 10 *p->n   + 2*p->mmax + 2;
        p->liwar = p->n;

/*
	ae = g_new(double, p->m*p->n);
	ane = g_new(double, p->m*p->n);
*/
	p->a_type = g_new(SolverConstraintType, p->m);
	p->a = g_new(double, p->mmax*p->n);
	p->b = g_new(double, p->m);
	p->d = g_new(double, p->n);
	p->c = g_new(double, p->n*p->n);	
	p->x = g_new(double, p->n);
	p->u = g_new(double, p->m + 2*p->n);
	p->war = g_new(double, p->lwar);
	p->iwar = g_new(int, p->liwar);
        p->eps1 = 1.11e-16; 
        p->iout = 0;
        p->ifail = 0;
        p->iprint = 0;

	return (SolverProgram)p;
}

static void
w_qld_delete (SolverProgram program)
{
	qp_qld_program *p  = (qp_qld_program*)program;

        printf ("w_qld_delete\n");

	g_free(p->a_type);
	g_free(p->a);
	g_free(p->b);
	g_free(p->c);
	g_free(p->d);
	g_free(p->x);
	g_free(p->u);
	g_free(p->war);
	g_free(p->iwar);
}

static void
w_qld_set_maxim (SolverProgram program)
{
	qp_qld_program *p  = (qp_qld_program*)program;
        printf ("w_qp_set_maxim\n");

	p->b_max = 1;
}

static void
w_qld_set_minim (SolverProgram program)
{
	qp_qld_program *p  = (qp_qld_program*)program;
        printf ("w_qp_set_minim\n");
	p->b_max = 0;
}

static void
w_qld_set_obj_fn (SolverProgram program, int col, gnm_float value)
{
	qp_qld_program *p  = (qp_qld_program*)program;

        printf ("w_qld_set_obj_fn %d, %" GNM_FORMAT_g "\n", col, value);
	assert(col<p->n);
	p->d[col] = value;
}

static void
w_qld_set_obj_mat_fn (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qp_qld_program *p  = (qp_qld_program*)program;

	printf(" w_qld_set_obj_mat_fn \n");
	assert (col < p->n);
	assert (row < p->n);	
	p->c[row*p->n + col] = value;
	p->c[col*p->n + row] = value;
}

static void
w_qld_set_constr_mat (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qp_qld_program *p  = (qp_qld_program*)program;

        printf ("w_qld_set_constr_mat %d, %d, %" GNM_FORMAT_g "\n", col, row, value);
	// now we don't know what type of constraint it is, filling both arrays
	assert (col < p->n);
	assert (row < p->m);
	p->a[row*p->n + col] = value;		
}

static void
w_qld_set_constr (SolverProgram program, int row,
		       SolverConstraintType type,
		       gnm_float value)
{
	qp_qld_program *p  = (qp_qld_program*)program;

        printf ("w_qld_set_constr %d, %d, %" GNM_FORMAT_g "\n", row, type, value);
	assert(row<p->m);

	p->b[row] = value;
	p->a_type[row] = type;
	if(type == SolverEQ) p->me++;

}

static void
w_qld_set_int (SolverProgram program, int col)
{
        printf ("w_qld_set_int %d\n", col);
}

static void
w_qld_set_bool (SolverProgram program, int col)
{
        printf ("w_qld_set_bool %d\n", col);
}

static SolverStatus
w_qld_solve (SolverProgram program)
{
	double *ae, *ane, *ptr;
	int i, j, ae_pos, ane_pos, linesize;

	qp_qld_program *p  = (qp_qld_program*)program;
        printf ("w_qld_solve\n");

	linesize = p->n*sizeof(double);
	//ae = g_new(double, p->me *p->n );
	//ane = g_new(double, (p->m - p->me) * p->n);
	ae = g_new(double, p->m * p->n);
	ane = g_new(double, p->m * p->n);

	ae_pos = ane_pos = 0;

	for(i=0;i<p->m;i++) {
		switch(p->a_type[i]) {
			case SolverEQ: 
				memcpy(ae+ ae_pos*p->n, p->a + i*p->n, linesize);	
				ae_pos++;
				break;
			case SolverLE:
				ptr = ane+ ane_pos*p->n;

				memcpy(ptr, p->a + i*p->n, linesize);	
				ane_pos++;
				for(j=0;j<p->n;j++) 
					ptr[j] = -ptr[j];
					
				break;
			case SolverGE:
				memcpy(ane+ ane_pos*p->n, p->a + i*p->n, linesize);	
				ane_pos++;
				p->b[i] = -p->b[i];
				break;
			case SolverINT:
			case SolverBOOL:
				printf(" SolverINT or SolverBOOL constraint type occured - HEEELP! \n");
				break;
		}
	}
					
	for(i=0;i<p->me;i++) 
		memcpy(p->a + i*p->n, ae + i*p->n, linesize);
		
	for(i=0;i<p->m-p->me;i++) 
		memcpy(p->a + (p->me+i)*p->n, ane + i*p->n, linesize);

	p->xl = g_new(double, p->n);
	p->xu = g_new(double, p->n);

	for(i=0;i<p->n;i++) {
		p->xl[i] = -1.0e+20;
		p->xu[i] =  1.0e+20;
	}

	if(p->b_max == 1) {
		for(i=0;i<p->n;i++) {
			p->d[i] = - p->d[i];			
			for(j=0;j<p->n;j++) {
				p->c[i*p->n+j] = -p->c[i*p->n+j] ;
			}
		}
	}


	p->iwar[0] = 1;		// this indicates the 0.5x'*C*x situation
        ql0001_(&p->m, &p->me, &p->mmax, &p->n, &p->nmax, &p->mnn, p->c, p->d, p->a, p->b, p->xl, p->xu, p->x, p->u, &p->iout, &p->ifail, &p->iprint, p->war, &p->lwar, p->iwar, &p->liwar, &p->eps1);

	if (ae != NULL) g_free(ae);
	if (ane != NULL) g_free(ane);
	g_free(p->xl);
	g_free(p->xu);

	switch (p->ifail) {
		case 0: return SolverOptimal;
		case 1: return SolverMaxIterExc;
		case 2: return SolverFailure;
		case 5: return SolverFailure;
		default:
			if(p->ifail > 10) return SolverInfeasible;	
			break;
	}
/*
	else if (ifail==1)
    Scierror(24,"qld: Too many iterations (more than %d) \r\n", 40*(n+m));
  else if (ifail==2)
    Scierror(24,"qld: Accuracy insufficient to statify convergence criterion.\r\n");
  else if (ifail==5)
    Scierror(999,"qld: Length of working array is too short.\r\n");
 else if (ifail>10)
    Scierror(999,"qld: The constraints are inconsistent.\r\n");
*/

}

static gnm_float
w_qld_get_solution (SolverProgram program, int col)
{
        printf ("w_qld_get_solution %d\n", col);

	return ((qp_qld_program*)program)->x[col];
}

static gnm_float
w_qld_get_value_of_obj_fn (SolverProgram program)
{
        printf ("w_qld_get_value_of_obj_fn\n");
	return 0;
}

static gnm_float
w_qld_get_dual (SolverProgram program, int row)
{
        printf ("w_qld_get_dual %d\n", row);
	return 0;
}

static int
w_qld_solver_lp_get_iterations (SolverProgram p)
{
	return 0;
}

static gboolean
w_qld_set_option (SolverProgram program, SolverOptionType option,
		   const gboolean *b_value,
		   const gnm_float *f_value, const int *i_value)
{
        printf ("w_qld_set_option %d\n", option);
        return FALSE;
}

static void
w_qld_print (SolverProgram program)
{
        printf ("w_qld_print\n");
}


/* ------------------------------------------------------------------------- */

/*
 * Solver's API wrappings for QP .
 *
 * Package:    OOQP
 * Version:    0
 * License:    GPL
 * Homepage:   http://www.cs.wisc.edu/~swright/ooqp/ 
 * Algorithm:  interior point method for convex quadratic programming
 *
 */

static SolverProgram
w_ooqp_init (const SolverParameters *param)
{
	qp_ooqp_program *p = g_new (qp_ooqp_program, 1);

        printf ("w_ooqp_init\n");
	assert(param->n_constraints == param->n_total_constraints);

	p->myz = param->n_constraints;
 	p->nx = param->n_variables;

	p->my = p->mz = 0;

	p->a_type = g_new(SolverConstraintType, p->myz);
	p->curA = g_new(double, p->nx);
	p->exp_row = -1;

	p->c = g_new(double, p->nx);
	p->Q = g_new(double, p->nx*p->nx);	

	p->A = g_new(double, p->myz*p->nx);
	p->bA = g_new(double, p->myz);
	p->C = g_new(double, p->myz*p->nx);
	p->clow = g_new(double, p->myz);
	p->iclow = g_new(char, p->myz);
	p->cupp = g_new(double, p->myz);
	p->icupp = g_new(char, p->myz);

	p->xlow = g_new(double, p->nx);
	p->ixlow = g_new(char, p->nx);
	p->xupp = g_new(double, p->nx);
	p->ixupp = g_new(char, p->nx);

	return (SolverProgram)p;
}

static void
w_ooqp_delete (SolverProgram program)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;

        printf ("w_ooqp_delete\n");

	g_free(p->curA);
	g_free(p->a_type);

	g_free(p->c);
	g_free(p->Q);	
	g_free(p->A);
	g_free(p->bA);
	g_free(p->C);
	g_free(p->clow);
	g_free(p->iclow);
	g_free(p->cupp);
	g_free(p->icupp);

	g_free(p->xlow);
	g_free(p->ixlow);
	g_free(p->xupp);
	g_free(p->ixupp);

	g_free(p->x); 
	g_free(p->gamma); 
	g_free(p->phi); 
	g_free(p->y);
	g_free(p->z); 
	g_free(p->lambda);
	g_free(p->pi);
	
	g_free(p);
}

static void
w_ooqp_set_maxim (SolverProgram program)
{
        printf ("w_qp_set_maxim\n");
	((qp_ooqp_program*)program)->min = 0;
	
}

static void
w_ooqp_set_minim (SolverProgram program)
{
        printf ("w_qp_set_minim\n");
	((qp_ooqp_program*)program)->min = 1;
}

static void
w_ooqp_set_obj_fn (SolverProgram program, int col, gnm_float value)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;

        printf ("w_ooqp_set_obj_fn %d, %" GNM_FORMAT_g "\n", col, value);
	assert(col<p->nx);
	p->c[col] = value;
}

static void
w_ooqp_set_obj_mat_fn (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;

	printf(" w_ooqp_set_obj_mat_fn \n");
	assert (col < p->nx);
	assert (row < p->nx);	
	p->Q[row*p->nx + col] = value;
	p->Q[col*p->nx + row] = value;
//   	a symmetric matrix
}

static void
w_ooqp_set_constr_mat (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;

        printf ("w_ooqp_set_constr_mat %d, %d, %" GNM_FORMAT_g "\n", col, row, value);
	// now we don't know what type of constraint it is, filling both arrays
	if(p->exp_row == -1) {
		assert (row < p->myz);
		p->exp_row = row;
	} else {
		assert (row == p->exp_row);
	}
	assert (col < p->nx);
	assert (0 <= col);

	p->curA[col] = value;
}

static void
w_ooqp_set_constr (SolverProgram program, int row,
		       SolverConstraintType type,
		       gnm_float value)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;
	int i;

        printf ("w_ooqp_set_constr %d, %d, %" GNM_FORMAT_g "\n", row, type, value);
	assert(row == p->exp_row);

	p->a_type[row] = type;

	switch(type) {
		case SolverLE:
			memcpy(p->C + p->mz*p->nx, p->curA, p->nx * sizeof(double));
			p->cupp[p->mz] = value;
			p->icupp[p->mz] = 1;
			p->clow[p->mz] = 0.0;
			p->iclow[p->mz] = 0;
			p->mz++;
			break;
		case SolverGE:
			memcpy(p->C + p->mz*p->nx, p->curA, p->nx * sizeof(double));
			p->clow[p->mz] = value;
			p->iclow[p->mz] = 1;
			p->cupp[p->mz] = 0.0;
			p->icupp[p->mz] = 0;
			p->mz++;
			break;
		case SolverEQ:
			memcpy(p->A + p->my*p->nx, p->curA, p->nx * sizeof(double)); 
			p->bA[p->my] = value;
			p->my++;
			break;
		case SolverINT:
			printf(" This solver doesn't support INT constraints. \n");
			break;
		case SolverBOOL:
			printf(" This solver doesn't support BOOL constraints. \n");
			break;
	}
	for(i=0;i<p->nx;i++) p->curA[i] = 0.0;
	p->exp_row = -1;
}

static void
w_ooqp_set_int (SolverProgram program, int col)
{
        printf ("w_ooqp_set_int %d\n", col);
}

static void
w_ooqp_set_bool (SolverProgram program, int col)
{
        printf ("w_ooqp_set_bool %d\n", col);
}

static SolverStatus
w_ooqp_solve (SolverProgram program)
{
	int i, ierr;

	qp_ooqp_program *p  = (qp_ooqp_program*)program;

        printf ("w_ooqp_solve\n");

	p->x = g_new(double, p->nx);
	p->gamma = g_new(double, p->nx);
	p->phi = g_new(double, p->nx);
	p->y = g_new(double, p->my);
	p->z = g_new(double, p->mz);
	p->lambda = g_new(double, p->mz);
	p->pi = g_new(double, p->mz);

	memset (p->ixlow, 0, p->nx * sizeof(char));
	memset (p->ixupp, 0, p->nx * sizeof(char));

	for(i=0;i<p->nx;i++) {
		p->xlow[i] = 0.0;
		p->xupp[i] = 0.0;
	}

	if(!p->min) {
		for(i=0;i<p->nx;i++) p->c[i] = - p->c[i];
		for(i=0;i<p->nx*p->nx;i++) p->Q[i] = - p->Q[i];
	}

	qpsolvede(p->c, p->nx, p->Q, 
		p->xlow, p->ixlow, p->xupp, p->ixupp,
                 p->A, p->my, p->bA,
                 p->C, p->mz, p->clow, p->iclow, p->cupp, p->icupp,

                 p->x, p->gamma, p->phi,
                 p->y,
                 p->z, p->lambda, p->pi, 
		 &p->ObjVal, 0, &ierr);

	switch (ierr) {
		case 0: return SolverOptimal;
		case 1: return SolverFailure;
		case 2: return SolverMaxIterExc;
		case 3: return SolverInfeasible;
		default: return SolverFailure;
/*
	SolverRunning, SolverOptimal, SolverUnbounded, SolverInfeasible,
	SolverFailure, SolverMaxIterExc, SolverMaxTimeExc
*/
	}
}

static gnm_float
w_ooqp_get_solution (SolverProgram program, int col)
{
        printf ("w_ooqp_get_solution %d\n", col);

	return ((qp_ooqp_program*)program)->x[col];
}

static gnm_float
w_ooqp_get_value_of_obj_fn (SolverProgram program)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;
        printf ("w_ooqp_get_value_of_obj_fn\n");

	return p->ObjVal;
}

static gnm_float
w_ooqp_get_dual (SolverProgram program, int row)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;
	int i, count;

        printf ("w_ooqp_get_dual %d\n", row);
	count = 0;
	switch(p->a_type[row]) {
		case SolverLE:
		case SolverGE:
			for(i=0;i<row;i++) {
				if(p->a_type[i] == SolverLE || p->a_type[i] == SolverGE) 
					count++;
			}
			return p->z[count];	

		case SolverEQ: 
			for(i=0;i<row;i++) {
				if(p->a_type[i] == SolverEQ) count++;
			}
			return p->y[count];

		default: 
			printf(" Cannot handle this the type of %d. constraint. \n", row);
			break;
	}	
	return 0;
}

static int
w_ooqp_solver_lp_get_iterations (SolverProgram program)
{
	qp_ooqp_program *p  = (qp_ooqp_program*)program;
	return 0;
}

static gboolean
w_ooqp_set_option (SolverProgram program, SolverOptionType option,
		   const gboolean *b_value,
		   const gnm_float *f_value, const int *i_value)
{
        printf ("w_ooqp_set_option %d\n", option);
        return FALSE;
}

static void
w_ooqp_print (SolverProgram program)
{
        printf ("w_ooqp_print\n");
}



/* ------------------------------------------------------------------------- */

/*
 * Solver's API wrappings for QP .
 *
 * Package:    COPLQP
 * Version:    0
 * License:    own 
 * Homepage:   http://www.stanford.edu/~yyye/Col.html
 * Algorithm:  interior point method for convex quadratic programming
 *
 */


static SolverProgram
w_coplqp_init (const SolverParameters *param)
{
	qpdat *p = g_new (qpdat, 1);
	int i;

        printf ("w_coplqp_init\n");
	assert(param->n_constraints == param->n_total_constraints);

	memset(p, 0, sizeof(qpdat));
	setParams(&p->par);

	p->m_intSta = 0;
	p->m_intRow = param->n_constraints;
 	p->m_intCol = p->m_colUse = param->n_variables;

	// initialize structure for columns
	p->col = g_new (optrow, p->m_intCol);
	memset(p->col, 0, p->m_intCol * sizeof(optrow));
	for(i=0;i<p->m_intCol;i++) {
		/*p->col[i].upp = p->par.q_dblUpp;        // TODO
		p->col[i].low = -p->par.q_dblUpp;        
		*/

		p->col[i].upp = 1000;        
		p->col[i].low = -1000;        	// i temporarily inserted some boundaries, so the solver can get some results.
						// without any bounds it cannot find a solution mostly
	}

	// initialize the matrix for objective function
	p->qh = g_new (symatx, 1);
	p->qh->ncol = p->m_intCol;
	p->qh->diag = g_new(double, p->m_intCol);
	for(i=0;i<p->m_intCol;i++) 
		p->qh->diag[i] = 0.0;

	p->qh->cols = g_new(array, p->m_intCol);
	memset(p->qh->cols, 0, p->m_intCol * sizeof(array));
	
	// condition types
	p->rowtype = g_new(char, p->m_intRow);

	// right side of the conditions
	p->b = g_new(double, p->m_intRow);
	p->r = g_new(double, p->m_intRow);

	// boundaries for the variables
	p->m_intNlo = p->m_intCol;
	p->m_intNup = p->m_intCol;

	return (SolverProgram)p;
}

static void
w_coplqp_delete (SolverProgram program)
{
	qpdat *p  = (qpdat*)program;

        printf ("w_coplqp_delete\n");

	FreeAll(p);
}

static void
w_coplqp_set_maxim (SolverProgram program)
{
        printf ("w_qp_set_maxim\n");
	((qpdat*)program)->par.q_intMin = false;
}

static void
w_coplqp_set_minim (SolverProgram program)
{
        printf ("w_qp_set_minim\n");
	((qpdat*)program)->par.q_intMin = true;
}

static void
w_coplqp_set_obj_fn (SolverProgram program, int col, gnm_float value)
{
	qpdat *p  = (qpdat*)program;

        printf ("w_coplqp_set_obj_fn %d, %" GNM_FORMAT_g "\n", col, value);
	assert(col<p->m_intCol);
	p->col[col].rhs = value;
}

static void
w_coplqp_set_obj_mat_fn (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qpdat *p  = (qpdat*)program;

	printf(" w_coplqp_set_obj_mat_fn \n");
	assert (col < p->m_intCol);
	assert (row < p->m_intCol);	
	if(value == 0.0) return;

	// only fill the half under the diagonal
	if(col > row) {
		int s = col;
		col = row;
		row = s;
	}
	
	if(col == row) {
		p->ndia++;
		p->qh->diag[col] = value;
	} else {
		array *acol = p->qh->cols + col;
		int pos = acol->nn0;

		p->m_intQnz++;
	
		acol->nn0++; 
		acol->ja = (int*)realloc(acol->ja, acol->nn0 * sizeof(int));	
		acol->an = (double*)realloc(acol->an, 
				acol->nn0 * sizeof(double));
		acol->ja[pos] = row;
		acol->an[pos] = value;
	}
}

static void
w_coplqp_set_constr_mat (SolverProgram program, int col, int row,
			   gnm_float value)
{
	qpdat *p  = (qpdat*)program;
	optrow *acol;
	int pos;

	if(value == 0.0) return;
	p->m_intAnz++;

        printf ("w_coplqp_set_constr_mat %d, %d, %" GNM_FORMAT_g "\n", col, row, value);
	// now we don't know what type of constraint it is, filling both arrays
	acol = p->col + col;
	pos = acol->nn0++;
	
	acol->ja = (int*)realloc(acol->ja, acol->nn0 * sizeof(int));
	acol->an = (double*)realloc(acol->an, acol->nn0 * sizeof(double));
	acol->ja[pos] = row;	
	acol->an[pos] = value;
}

static void
w_coplqp_set_constr (SolverProgram program, int row,
		       SolverConstraintType type,
		       gnm_float value)
{
	qpdat *p  = (qpdat*)program;

        printf ("w_coplqp_set_constr %d, %d, %" GNM_FORMAT_g "\n", row, type, value);

	p->b[row] = value;
	p->r[row] = value;

	switch(type) {
		case SolverLE:
			p->m_intUne++;
			p->rowtype[row] = 'L';
			break;
		case SolverGE:
			p->m_intUne++;
			p->rowtype[row] = 'G';
			break;
		case SolverEQ:
			p->rowtype[row] = 'E';
			break;
		case SolverINT:
			printf(" This solver doesn't support INT constraints. \n");
			break;
		case SolverBOOL:
			printf(" This solver doesn't support BOOL constraints. \n");
			break;
	}
}

static void
w_coplqp_set_int (SolverProgram program, int col)
{
        printf ("w_coplqp_set_int %d\n", col);
}

static void
w_coplqp_set_bool (SolverProgram program, int col)
{
        printf ("w_coplqp_set_bool %d\n", col);
}

static SolverStatus
w_coplqp_solve (SolverProgram program)
{
	qpdat *p  = (qpdat*)program;

        printf ("w_coplqp_solve\n");

	p->m_intNsk = p->m_intUne;

	SolveQP(p);

	switch(p->m_intSta) {
    		case OPT_UNSOL:
			return SolverFailure;	
		case OPT_OK:
			return SolverOptimal;
    		case OPT_FOUND:
			return SolverOptimal;
		case INF_PROB:
			return SolverInfeasible;
		case UNB_PROB:
			return SolverUnbounded;
		case EXC_ITER:
			return SolverMaxIterExc;
    		case FEA_PROB:
		case UNK_SOL:
		case UNK_PROB:
		case NUM_PROB:
			return SolverFailure;
	}	
}

static gnm_float
w_coplqp_get_solution (SolverProgram program, int col)
{
        printf ("w_coplqp_get_solution %d\n", col);

	return ((qpdat*)program)->x[col];
}

static gnm_float
w_coplqp_get_value_of_obj_fn (SolverProgram program)
{
	qpdat *p  = (qpdat*)program;
        printf ("w_coplqp_get_value_of_obj_fn\n");

	return p->obj; 
}

static gnm_float
w_coplqp_get_dual (SolverProgram program, int row)
{
	qpdat *p  = (qpdat*)program;

        printf ("w_coplqp_get_dual %d\n", row);
	return p->rmem[row];	
}

static int
w_coplqp_solver_lp_get_iterations (SolverProgram program)
{
	qpdat *p  = (qpdat*)program;
	return 0;
}

static gboolean
w_coplqp_set_option (SolverProgram program, SolverOptionType option,
		   const gboolean *b_value,
		   const gnm_float *f_value, const int *i_value)
{
        printf ("w_coplqp_set_option %d\n", option);
        return FALSE;
}

static void
w_coplqp_print (SolverProgram program)
{
        printf ("w_coplqp_print\n");
}





/* ------------------------------------------------------------------------- */

/*
 * This array contains the linear programming algorithms available.
 * Each algorithm should implement the API defined by the
 * SolverLPAlgorithm data structure.  The algorithms should be able to
 * do MILP as well.  Feel free to add new algorithms.
 */
const SolverLPAlgorithm lp_algorithm [] = {
        {
	        NULL,
		(solver_init_fn*)                w_lp_solve_init,
		(solver_remove_fn*)              w_lp_solve_delete_lp,
		(solver_lp_set_obj_fn*)          w_lp_solve_set_obj_fn,
		(solver_qp_set_obj_mat_fn*)	 w_lp_set_obj_mat_fn_dummy, 
		(solver_lp_set_constr_mat_fn*)   w_lp_solve_set_constr_mat,
		(solver_lp_set_constr_fn*)       w_lp_solve_set_constr,
		(solver_lp_set_maxim_fn*)        w_lp_solve_set_maxim,
		(solver_lp_set_minim_fn*)        w_lp_solve_set_minim,
		(solver_lp_set_int_fn*)          w_lp_solve_set_int,
		(solver_lp_set_bool_fn*)         w_lp_solve_set_bool,
		(solver_lp_solve_fn*)            w_lp_solve_solve,
		(solver_lp_get_obj_fn_value_fn*) w_lp_solve_get_value_of_obj_fn,
		(solver_lp_get_obj_fn_var_fn*)   w_lp_solve_get_solution,
		(solver_lp_get_shadow_prize_fn*) w_lp_solve_get_dual,
		(solver_lp_get_iterations_fn*)   w_lp_solve_get_iterations,
		(solver_lp_set_option_fn*)       w_lp_solve_set_option,
		(solver_lp_print_fn*)            w_lp_solve_print_lp
	},

#if __HAVE_GLPK__
        {
	        NULL,
		(solver_init_fn*)                w_glpk_init,
		(solver_remove_fn*)              w_glpk_delete_lp,
		(solver_lp_set_obj_fn*)          w_glpk_set_obj_fn,
		(solver_qp_set_obj_mat_fn*)	 w_lp_set_obj_mat_fn_dummy, 
		(solver_lp_set_constr_mat_fn*)   w_glpk_set_constr_mat,
		(solver_lp_set_constr_fn*)       w_glpk_set_constr,
		(solver_lp_set_maxim_fn*)        w_glpk_set_maxim,
		(solver_lp_set_minim_fn*)        w_glpk_set_minim,
		(solver_lp_set_int_fn*)          w_glpk_set_int,
		(solver_lp_set_bool_fn*)         w_glpk_set_bool,
		(solver_lp_solve_fn*)            w_glpk_simplex_solve,
		(solver_lp_get_obj_fn_value_fn*) w_glpk_get_value_of_obj_fn,
		(solver_lp_get_obj_fn_var_fn*)   w_glpk_get_solution,
		(solver_lp_get_shadow_prize_fn*) w_glpk_get_dual,
		(solver_lp_get_iterations_fn*)   w_glpk_get_iterations,
		(solver_lp_set_option_fn*)       w_glpk_set_option,
		(solver_lp_print_fn*)            w_glpk_print_lp
	},
#endif


        {
	        NULL,
		(solver_init_fn*)                w_qld_init,
		(solver_remove_fn*)              w_qld_delete,
		(solver_lp_set_obj_fn*)          w_qld_set_obj_fn,
		(solver_qp_set_obj_mat_fn*) 	 w_qld_set_obj_mat_fn,
		(solver_lp_set_constr_mat_fn*)   w_qld_set_constr_mat,
		(solver_lp_set_constr_fn*)       w_qld_set_constr,
		(solver_lp_set_maxim_fn*)        w_qld_set_maxim,
		(solver_lp_set_minim_fn*)        w_qld_set_minim,
		(solver_lp_set_int_fn*)          w_qld_set_int,
		(solver_lp_set_bool_fn*)         w_qld_set_bool,
		(solver_lp_solve_fn*)            w_qld_solve,
		(solver_lp_get_obj_fn_value_fn*) w_qld_get_value_of_obj_fn,
		(solver_lp_get_obj_fn_var_fn*)   w_qld_get_solution,
		(solver_lp_get_shadow_prize_fn*) w_qld_get_dual,
		(solver_lp_get_iterations_fn *)  w_qld_solver_lp_get_iterations,
		(solver_lp_set_option_fn*)       w_qld_set_option,
		(solver_lp_print_fn*)            w_qld_print
	},

        {
	        NULL,
		(solver_init_fn*)                w_ooqp_init,
		(solver_remove_fn*)              w_ooqp_delete,
		(solver_lp_set_obj_fn*)          w_ooqp_set_obj_fn,
		(solver_qp_set_obj_mat_fn*) 	 w_ooqp_set_obj_mat_fn,
		(solver_lp_set_constr_mat_fn*)   w_ooqp_set_constr_mat,
		(solver_lp_set_constr_fn*)       w_ooqp_set_constr,
		(solver_lp_set_maxim_fn*)        w_ooqp_set_maxim,
		(solver_lp_set_minim_fn*)        w_ooqp_set_minim,
		(solver_lp_set_int_fn*)          w_ooqp_set_int,
		(solver_lp_set_bool_fn*)         w_ooqp_set_bool,
		(solver_lp_solve_fn*)            w_ooqp_solve,
		(solver_lp_get_obj_fn_value_fn*) w_ooqp_get_value_of_obj_fn,
		(solver_lp_get_obj_fn_var_fn*)   w_ooqp_get_solution,
		(solver_lp_get_shadow_prize_fn*) w_ooqp_get_dual,
		(solver_lp_get_iterations_fn *)  w_ooqp_solver_lp_get_iterations,
		(solver_lp_set_option_fn*)       w_ooqp_set_option,
		(solver_lp_print_fn*)            w_ooqp_print
	},
        {
	        NULL,
		(solver_init_fn*)                w_coplqp_init,
		(solver_remove_fn*)              w_coplqp_delete,
		(solver_lp_set_obj_fn*)          w_coplqp_set_obj_fn,
		(solver_qp_set_obj_mat_fn*) 	 w_coplqp_set_obj_mat_fn,
		(solver_lp_set_constr_mat_fn*)   w_coplqp_set_constr_mat,
		(solver_lp_set_constr_fn*)       w_coplqp_set_constr,
		(solver_lp_set_maxim_fn*)        w_coplqp_set_maxim,
		(solver_lp_set_minim_fn*)        w_coplqp_set_minim,
		(solver_lp_set_int_fn*)          w_coplqp_set_int,
		(solver_lp_set_bool_fn*)         w_coplqp_set_bool,
		(solver_lp_solve_fn*)            w_coplqp_solve,
		(solver_lp_get_obj_fn_value_fn*) w_coplqp_get_value_of_obj_fn,
		(solver_lp_get_obj_fn_var_fn*)   w_coplqp_get_solution,
		(solver_lp_get_shadow_prize_fn*) w_coplqp_get_dual,
		(solver_lp_get_iterations_fn *)  w_coplqp_solver_lp_get_iterations,
		(solver_lp_set_option_fn*)       w_coplqp_set_option,
		(solver_lp_print_fn*)            w_coplqp_print
	},

	{ NULL }
};

